import java.awt.*;
import java.awt.image.BufferedImage;
import javax.swing.*;


/**
 * Static class for showing dialog windows.
 */
public class Dialogs {
	static Laf icApp;
	
	/**
	 * Shows the window for inputting window size.
	 * @param f Frame to make this dialog belong to.
	 * @param w Initial width.
	 * @param h Initial height.
	 * @return Dimension as specified by the user, or null if user canceled or closed the window.
	 */
	public static Dimension showSizeInputDialog(int w, int h){
		// create the dialog.
		SizeInputDialog s=new SizeInputDialog(icApp, w, h);
		if(s.hitOK)
			// user didn't cancel or close the window.
			return new Dimension(s.width, s.height);
		return null;
	}
	
	/**
	 * Shows the window for choosing features.
	 * @param f Frame this dialog belongs to.
	 * @param showAlignAndBlend true to show align and blend features, false otherwise.
	 * @return 3 element array, where 1st element is for alignment, 2nd is for blending,
	 * 			and 3rd for bilinear interpolation.
	 */
	public static boolean[] showFeaturesDialog(boolean showAlignAndBlend){
		FeatureSelectionDialog d=new FeatureSelectionDialog(icApp, showAlignAndBlend);
		boolean [] r={d.warpAlign && showAlignAndBlend, d.blend && showAlignAndBlend, d.biInterp};
		if(d.hitOK)
			return r;
		return null;
	}
	
	/**
	 * Shows dialog for mosaicing images.
	 * @param im Original image to mosaic onto.
	 * @return Mosaiced image or null if Calcel was pressed.
	 */
	public static BufferedImage showMosaicDialog(BufferedImage im){
		JDialog jd=new JDialog(icApp, "Mosaic Image", true);
		MosaicTargetPanel ip=new MosaicTargetPanel(im, (Laf)icApp, jd);
		jd.setContentPane(ip);
		jd.setResizable(true);
		jd.pack();
		jd.setLocation(icApp.getX()+(icApp.getWidth()-jd.getWidth())/2,
			icApp.getY()+(icApp.getHeight()-jd.getHeight())/2);
		jd.setVisible(true);
		if(ip.hitOK)
			return ip.getImage();
		return null;
	}
	
	/**
	 * Shows the progress bar window.
	 * @param utils Object which which contains the updated progress.
	 * @param total Total progress.
	 */
	public static void showProgressBar(Progressable utils, int total){
		new ProgressFrame(icApp, utils, total);
	}
	
	/**
	 * Shows the foreground extraction dialog.
	 * @param im Image to extract foreground from.
	 * @return Image with the background colors removed.
	 */
	public static BufferedImage showForegroundExtr(BufferedImage im){
		ForegroundExtractor f=new ForegroundExtractor(icApp, im);
		if(f.hitSetForeground){
			FgExtract fgx=new FgExtract(f);
			return (BufferedImage)fgx.runOperations();
		}
		return null;
	}
	
	/** Class for doing the foreground extraction and bringing up the progress dialog. */
	private static class FgExtract extends ProcessesLump{
		ForegroundExtractor f;
		
		public FgExtract(ForegroundExtractor f){this.f=f;}
		
		/* (non-Javadoc)
		 * @see ProcessesLump#precompute()
		 */
		protected boolean precompute() {
			return true;
		}

		/* (non-Javadoc)
		 * @see ProcessesLump#estimateProgress()
		 */
		protected int estimateProgress() {
			return f.estimateExtractForegroundProgress();
		}

		/* (non-Javadoc)
		 * @see ProcessesLump#exec()
		 */
		protected Object exec() {
			return f.extractForeground();
		}
	}
	
	/**
	 * Brings up the about dialog.
	 */
	public static void showAboutDialog(){
		new AboutDialog(icApp);
	}
	
	/**
	 * Brings up the help dialog.
	 */
	public static void showHelp(){
		((JPanel)HelpDialog.getHelpDialog(icApp).getContentPane()).grabFocus();
	}
}
