
/** Class for a min-heap modified for finding nodes inside the heap faster.*/

public class Heap {
	private HeapNode[] recs; // An array that will hold the heap
	private int[] locations;
	
	private int length;
	private int maxlength;
	private int im_width;
	private int im_height;
	
	/**
	 * Constructor for Heap
	 * @param im_height - height of image
	 * @param im_width - width of image
	 */
	public Heap(int im_height, int im_width){
		length = 0;
		maxlength=im_height*im_width;
		this.im_height = im_height;
		this.im_width = im_width;
		recs = new HeapNode[maxlength+1];
		locations = new int[maxlength];
		for(int i=0;i<maxlength;i++){
			locations[i] = 0;
		}
	}
	
	/**
	 * Adds an event to the heap.
	 * Required: The Heap isn't full.
	 *			 e exists 
	 * @param rec - current Heap
	 */
	public void enqueue(HeapNode rec){
		if(length<maxlength)
		{
			recs[++length]=rec;
			locations[rec.getY()*im_width+rec.getX()]=length;
			sortup(length);
		}else{
			System.out.println("Unable to add record, because the heap is full.\n");
		}
	}
	
	/**
	 * Remove and return the HeapNode to the Event object at
	 * the top of the heap.
	 * @return The pointer to the top Event object (that is to be executed next).
	 */
	public HeapNode dequeue(){
		HeapNode rec=recs[1];
		if(length==0) 
			return null;
		locations[recs[1].getY()*im_width+recs[1].getX()]=0;
		recs[1]=recs[length];
		if(length!=1)
			locations[recs[1].getY()*im_width+recs[1].getX()]=1;
		recs[length--]=null;
		sortdown(1);
		return rec;
	}
	
	/**
	 * Return the heap's length
	 * @return - length of heap
	 */
	public int getLength(){return length;}
	
	public HeapNode findNode(int r, int c){
		return recs[locations[r*im_width+c]];
	}
	
	public void delNode(int r, int c){
		int ind=locations[r*im_width+c];
		if(ind==0) return;

		recs[ind] = null;

		locations[r*im_width+c]=0;
		recs[ind]=recs[length];
		if(ind!=length)
			locations[recs[ind].getY()*im_width+recs[ind].getX()]=ind;
		recs[length--]=null;
		sortdown(ind);
	}
	
	public boolean isNotEmpty(){
		if(length>0)
			return true;
		return false;
	}
	
	/**
	 * Moves an Event at index 'ind' up the heap until its scheduled
	 * time is less than of its children.
	 * Required: eventQueue[ind] exists
	 *	     	 1<=ind<length of the array.
	 * @param ind - index of the Event object to move up.
	 */
	private void sortup(int ind){
		HeapNode rec;
		
		while(ind!=1 && recs[ind/2].getCost() > recs[ind].getCost()){
			// The time of the parent Event is greater than of its child,
			// so they need to be switched to maintain a sortedorder.
			
			rec=recs[ind];
			recs[ind]=recs[ind/2];
			locations[recs[ind].getY()*im_width+recs[ind].getX()]=ind;
			recs[ind/2]=rec;
			locations[recs[ind/2].getY()*im_width+recs[ind/2].getX()]=ind/2;
			ind=ind/2;
		}
	}
	
	/**
	 * Moves an Event at index 'parent' down the heap until
	 * its scheduled time is less than of its childrens'.
	 * Required: eventQueue[parent] exists
	 *	    	 1<=parent<length of the array.
	 * @param parent - index of the Event object to move down.
	 */
	private void sortdown(int parent){
		int child;
		HeapNode rec;
		/* Move the HeapNode object down the heap hiararchy, by switching it
		   with the child that is smallest of the two.*/
		for(child=parent*2; child<=length; parent=child,child=child*2){
			/* Unless this is the right most heapNode in the heap
				make 'child' be the index of the smaller
				record, to maintain proper order in the heap.*/
			if(child<length && recs[child].getCost()>recs[child+1].getCost())
				child=child+1;

			if(recs[parent].getCost()>recs[child].getCost()){
				/* The parent Event has greater time than the child
					so switch them. */
				rec=recs[child];
	            recs[child]=recs[parent];
				locations[recs[child].getY()*im_width+recs[child].getX()]=child;
	            recs[parent]=rec;
				locations[recs[parent].getY()*im_width+recs[parent].getX()]=parent;
			}else{
				break;
			}
		}
	}
}
