import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;

/** Panel holding the target image for mosaicing. */
class MosaicTargetPanel extends ImagePanel implements ActionListener {
	private Laf icApp;
	private JDialog frame, mosaicSourceFrame;
	private WarpSourcePanel mosaicSource;
	private SelectorPoint[] path;	// 4 target points.
	private SelectorPoint drag;		// point currently selected.
	private RegionPanel rp;			// The region holding the image.
	private LayoutManager bLayout;		// layout, needed to get dimension for the window.
	private XYPair rc1, rc2, rc3, rc4;	// 4 target points, needed here for multithreading.
	private boolean [] features;		// 3 features, warp-alignment, image-blending, & bilin-interpolation
	
	public boolean hitOK=false;

	/** Constructor:.
	 * @param file The path to the image file.
	 * @param icApp The main ImageCompositing application.
	 * @param frame The dialog this panel is in.
	 */
	public MosaicTargetPanel(BufferedImage bim, Laf icApp, JDialog frame) {
		this.icApp = icApp;
		this.frame=frame;

		// set up keyboard events to listen to arrows keys.
		InputMap im = getInputMap();
		ActionMap am = getActionMap();
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), "down");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_DOWN, 0), "down");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), "up");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_UP, 0), "up");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_LEFT, 0), "left");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "left");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_RIGHT, 0), "right");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "right");

		// event handlers for keys, perform nudges to the selectors.
		am.put("down", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.y++;
				repaint();
			}
		});
		am.put("up", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.y--;
				repaint();
			}
		});
		am.put("left", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.x--;
				repaint();
			}
		});
		am.put("right", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.x++;
				repaint();
			}
		});

		// create components of the panel.
		setLayout((bLayout = new BorderLayout()));
		
		JToolBar tb=new JToolBar();
		tb.setFloatable(false);
		add(tb, BorderLayout.NORTH);
		
		// Button for doing mosaicing.
		JButton b=new JButton("Add Image To Mosaic");
		b.setFocusable(false);
		b.addActionListener(this);
		tb.add(b);
		
		b=new JButton("Do Mosaic");
		b.setFocusable(false);
		b.addActionListener(this);
		tb.add(b);
		
		tb=new JToolBar();
		tb.setFloatable(false);
		add(tb, BorderLayout.SOUTH);
		
		// buttons for finishing mosaicing.
		b=new JButton("OK");
		b.setFocusable(false);
		b.addActionListener(this);
		tb.add(b);
		
		b=new JButton("Cancel");
		b.setFocusable(false);
		b.addActionListener(this);
		tb.add(b);
		
		// empty region for the user to use.
		rp = new RegionPanel();
		rp.setImage(bim);
		add(rp);

		SelectorPoint[] tmp ={
				new SelectorPoint(3, 3, 3, Color.BLUE),
				new SelectorPoint(3, 3, rp.getImageHeight() - 4, Color.RED),
				new SelectorPoint(3, rp.getImageWidth() - 4, 3, Color.GREEN),
				new SelectorPoint(3,rp.getImageWidth() - 4,rp.getImageHeight() - 4,Color.YELLOW)
		};
		path = tmp;
	}

	/**
	 * Overides parents method, since this panel doesn't actually hold an image.
	 * @return Image supposedly in this panel.
	 */
	public BufferedImage getImage() {
		if (rp == null)
			return null;
		return rp.getImage();
	}
	
	/** Puts image in the file into this panel.
	 * @param filename File path of the image.
	 */
	public boolean setImage(String fileName){
		return rp.setImage(fileName);
	}

	/** Puts an image into this panel.
	 * @param bimg Image.
	 */
	public void setImage(BufferedImage bimg){
		rp.setImage(bimg);
	}

	/** Paints image 'b' on top of the image in this panel.
	 * @param b Image to paint.
	 * @param x X coordinate to place the image at.
	 * @param y Y coordinate to place the image at.
	 */
	public void paintImage(BufferedImage b, int x, int y){ 
		rp.paintImage(b, x, y);
	}

	/** Returns a copy of this image.
	 * @return Copy of the image in this panel.
	 */
	public BufferedImage copyImage(){
		return rp.copyImage();
	}

	/** Returns image width.
	 * @return Image width.
	 */
	public int getImageWidth(){ return rp.getImageWidth(); }
	/** Returns image height.
	 * @return Image height.
	 */
	public int getImageHeight(){ return rp.getImageHeight(); }

	/**
	 * Handles the button click event.
	 * @param e Event object describing the event.
	 */
	public void actionPerformed(ActionEvent e) {
		if(e.getActionCommand().equals("Add Image To Mosaic")){
			int returnValue =icApp.fileChooser.showOpenDialog(icApp);
			if (returnValue == JFileChooser.APPROVE_OPTION) {
				String fileName =icApp.fileChooser.getSelectedFile().getAbsolutePath();
				if(mosaicSourceFrame!=null)
					mosaicSourceFrame.dispose();
				// source image.
				mosaicSource=new WarpSourcePanel(fileName, icApp);
				mosaicSourceFrame=new JDialog(frame, fileName);
				mosaicSourceFrame.setContentPane(mosaicSource);
				mosaicSourceFrame.pack();
				mosaicSourceFrame.setLocation(icApp.getX()+(icApp.getWidth()-mosaicSourceFrame.getWidth())/2,
					icApp.getY()+(icApp.getHeight()-mosaicSourceFrame.getHeight())/2);
				mosaicSourceFrame.setVisible(true);
				mosaicSourceFrame.addWindowListener(new WindowAdapter() {
					public void windowClosing(WindowEvent e) {
						mosaicSource=null;
						mosaicSourceFrame=null;
					}
				});
			}
		}else if(e.getActionCommand().equals("OK")){
			if(mosaicSourceFrame!=null)
				mosaicSourceFrame.dispose();
			hitOK=true;
			frame.dispose();
		}else if(e.getActionCommand().equals("Cancel")){
			if(mosaicSourceFrame!=null)
				mosaicSourceFrame.dispose();
			frame.dispose();
		}else if (mosaicSource != null) {
			// do mosaicing
			SelectorPoint [] sPath = mosaicSource.getPath();
			features=Dialogs.showFeaturesDialog(true);
			if(features==null)
				return;
			// Create the mosaic.
			rp.setImage(
				Paint.createMosaic(
					mosaicSource.getImage(),
					rp.getImage(),
					sPath[0].center,
					sPath[1].center,
					sPath[2].center,
					sPath[3].center,
					path[0].center,
					path[1].center,
					path[2].center,
					path[3].center, features));
			frame.pack();
			grabFocus();
		}
	}
	
	/**
	 * Overrides parent's method because this doesn't actually store the image.
	 * @param state State to set the image to.
	 */
	public void setState(int state){
		super.setState(state);
		rp.setState(state);
	}
	
	/**
	 * Overrides parent' method because this doesn't actually store the image.
	 */
	public BufferedImage getSelection(){
		return rp.getSelection();
	}

	/** Class for the area of this window holding the image. */
	private class RegionPanel extends ImagePanel {

		/** Event handler for mouse presses.
		 * @param e Event object describing the mouse event.
		 */
		public void mousePressed(MouseEvent e) {
			int x = e.getX(), y = e.getY();
			if (drag != null) {
				drag.selected = false;
				drag = null;
			}
			for (int i = 0; i < path.length; i++) {
				if (path[i].hitTest(x, y)) {
					drag = path[i];
					drag.selected = true;
					break;
				}
			}
			repaint();
		}

		/** Event handler for the mouse move event.
		 * @param e Object describing the event.
		 */
		public void mouseMoved(MouseEvent e) {
			int x = e.getX(), y = e.getY();
			for (int i = 0; i < path.length; i++) {
				if (path[i].hitTest(x, y)) {
					setCursor(new Cursor(Cursor.HAND_CURSOR));
					return;
				}
			}
			setCursor(Cursor.getDefaultCursor());
		}

		/** Event handler for the mouse drag event.
		 * @param e Object describing the event.
		 */
		public void mouseDragged(MouseEvent e) {
			if (drag == null)
				return;
			drag.center.x = e.getX();
			drag.center.y = e.getY();
			repaint();
		}
		
		/** Paints this component.
		 * @param g Graphics object to paint on.
		 */
		public void paintComponent(Graphics g) {
			super.paintComponent(g);
			for (int i = 0; path != null && i < path.length; i++)
				path[i].paint(g);
		}
	}
}
