import java.util.*;
import java.awt.event.MouseEvent;
import java.awt.image.*;

import javax.swing.JOptionPane;

/**
* Class that contains utilities for scissoring.
*/
public class Scissor implements ICOperation{
	private ImagePanel imgPanel;
	private BufferedImage image;
	private int [][] logzx;			// Zero crossings.
	private GradVector [][] grad; 	//every pixel has magnitude (index 0) and direction (index 1).

	private boolean closedPath=true;	// Is the selection path closed or not.
	private Vector seeds, path;			// Seeds and Shape objects making up the path.
	private int lastPathSeg;			// Index of the shape object with the path
										// segment to replace during live-wiring.
	private XYPair dest;				// Destination pixel, to draw a path from to the seed.
	double gradMax;
	HeapNode[][] pnts;	// Stores heap nodes that point to next heap nodes on their path to the seed.
	XYPair seed;		// Last seed point.
	
	/** Constructor:
	 * @param img Image to scissor.
	 * @param path Vector of Shape objects which will be drawn on the image panel.
	 * 2param imgPanel ImagePanel object holding the image to scissor.
	 */
	public Scissor(BufferedImage img, Vector path, ImagePanel imgPanel){
		this.imgPanel=imgPanel;
		this.path=path;
		seeds=new Vector();
		image=img;
	}
	
	/** Initilizes the data structures for scissoring to work.
	 * @return true if everything initialized OK, false otherwise.
	 */
	public boolean initialize(){
		RunInit init=new RunInit();
		if(init.runOperations()==null)
			return false;
		
		// Find maximum gradient magnitude.
		gradMax=0;
		for(int i=0; i<grad.length; i++){
			for(int j=0; j<grad[0].length; j++){
				if(grad[i][j].mag>gradMax) gradMax=grad[i][j].mag;
			}
		}
		
		return true;
	}

	/** Returns path from seed to destination pixel.
	 * @param dest Pixel to look for a path from.
	 * @return Vector holding XYPairs that correspond to the path between seed and dest.
	 */
	public Vector getPath(XYPair dest) {

		int r1 = seed.y, c1 = seed.x, r2 = dest.y, c2 = dest.x;

		HeapNode q;

		int cR = r2, cC = c2;
		Vector path = new Vector();
		//Trace the lowest cost path backwards from second pixel to first.
		while (true) {
			q = pnts[cR][cC];
			cR = q.getY();
			cC = q.getX();
			path.add(new XYPair(cC, cR));
			
			if (cR == r1 && cC == c1)
				break;
		}
		return path;
	}
	
	/** Compute shortest paths from all pixels to seed.
	 * @param seed Seed pixel.
	 */
	public void computePaths(XYPair seed){
		int r1 = seed.y, c1 = seed.x;
		int im_height = logzx.length, im_width = logzx[0].length;
		double tmp;
		
		this.seed=seed;

		HeapNode q, r;

		// tells if a heap node has been expanded.
		boolean[][] expd = new boolean[im_height][im_width];
		
		pnts = new HeapNode[im_height][im_width];

		// initilize the arrays.
		for (int i = 0; i < im_height; i++) {
			for (int j = 0; j < im_width; j++) {
				expd[i][j] = false;
				pnts[i][j] = null;
			}
		}
		pnts[r1][c1] = new HeapNode(r1, c1, 0);
		Heap pList = new Heap(im_height, im_width);
		pList.enqueue(new HeapNode(r1, c1, 0.0));

		// Get all the minimum paths to the seed.
		while (pList.isNotEmpty()) {
			q = pList.dequeue();
			expd[q.getY()][q.getX()] = true;

			// Go through all the neighbors of this pixel.
			for (int i = Math.max(0, q.getY() - 1);
				i < Math.min(im_height, q.getY() + 2);
				i++) {
				for (int j = Math.max(0, q.getX() - 1);
					j < Math.min(im_width, q.getX() + 2);
					j++) {
					if (!expd[i][j]) {
						// pixel not expanded yet.
						tmp =q.getCost()+ l_C(q.getY(),q.getX(),i,j);

						r = pList.findNode(i, j);

						if (r != null && tmp < r.getCost()) {
							// thispixel's path cost will change so remove it temporarily
							// it'll be added in the next if statement.
							pList.delNode(i, j);
							r = null;
						}
						if (r == null) {
							// create a heap node for this neighbor.
							pList.enqueue(new HeapNode(i, j, tmp));
							if (pnts[i][j] != null)
								pnts[i][j] = null;
							// set it to point to 'q'.
							pnts[i][j] = new HeapNode(q.getY(), q.getX(), 0);
						}
					}
				}
			}
			q = null;
		}
		pList = null;
	}

	/** Cost function.
	 * @param r1 y coordinate of 1st pixel.
	 * @param c1 x coordinate of 1st pixel.
	 * @param r2 y coordinate of 2nd pixel.
	 * @param c2 x coordinate of 2nd pixel.
	 * @return cost of the edge between 1st pixel and 2nd.
	 */
	private double l_C(int r1, int c1, int r2, int c2) {

		double scaler = 1.0;

		// favors horizontal and vertical neighbors.
		if (((r1 - r2) * (c1 - c2)) == 0)
			scaler = 1.0 / Math.sqrt(2);
		return (0.43 * f_Z(r2, c2))
			+ (0.14 * f_D(r1, c1, r2, c2))
			+ (0.43 * f_G(r2, c2));
	}

	/** Compute function fZ for cost function.
	 * @param r y coordinate of the pixel.
	 * @param c x coordinate of the pixel.
	 * @return fZ term.
	 */
	private double f_Z(int r, int c) {
		if (logzx[r][c] == 255)
			return 0.0;

		return 1.0;
	}

	/** Compute function fG for low cost function.
	 * @param r y coordinate of the pixel.
	 * @param c x coordinate of the pixel.
	 * @return fG term.
	 */
	private double f_G(int r, int c) {
		return (1.0 - (grad[r][c].mag / gradMax));
	}

	/** Compute function fD for low cost function.
	 * @param r1 y coordinate of 1st pixel.
	 * @param c1 x coordinate of 1st pixel.
	 * @param r2 y coordinate of 2nd pixel.
	 * @param c2 x coordinate of 2nd pixel.
	 * @return fD term.
	 */
	private double f_D(int r1, int c1, int r2, int c2) {

		int dx, dy;

		double tmp;
		double Ix_p = -grad[r1][c1].hor, Iy_p = grad[r1][c1].vert;
		double Ix_q = -grad[r2][c2].hor, Iy_q = grad[r2][c2].vert;
		if (((double) (r2 - r1) * Iy_p + (double) (c2 - c1) * Ix_p) >= 0.0) {
			dy = r2 - r1;
			dx = c2 - c1;
		} else {
			dy = r1 - r2;
			dx = c1 - c2;
		}
		return (Math.PI / 2)
			+ ((2.0 / (3.0 * Math.PI))
			* (Math.acos((Iy_p * dy + Ix_p * dx) / 2) + Math.acos((Iy_q * dy + Ix_q * dx) / 2)));
	}
	
	/** Finds an edge pixel closest to pixel 'xy'.
	 * @param xy Pixel to find closest edge pixel for.
	 * @return Edge pixel closest to 'xy'.
	 */
	public XYPair snapToEdge(XYPair xy){
		XYPair ret=new XYPair(xy.x, xy.y);
		
		int size=1, w=logzx[0].length, h=logzx.length;
		boolean foundEdge=false;
		// expand the area looking for an edge pixel.
		while(true){
			// See if there's an edge pixel on topmost row.
			foundEdge=isThereEdgeH(Math.max(0, xy.x-size/2), Math.min(xy.x+size/2+1, w),
				Math.max(0, xy.y-size/2));
			if(foundEdge) break;
			
			// See if there's an edge pixel on bottommost row.
			foundEdge=isThereEdgeH(Math.max(0, xy.x-size/2), Math.min(xy.x+size/2+1, w),
				Math.min(xy.y+size/2, h-1));
			if(foundEdge) break;

			// See if there's an edge pixel on leftmost column.
			foundEdge=isThereEdgeV(Math.max(0, xy.x-size/2), Math.max(0, xy.y-size/2),
				Math.min(h, xy.y+size/2+1));	
			if(foundEdge) break;
			
			// See if there's an edge pixel on rightmost column.
			foundEdge=isThereEdgeV(Math.min(w-1, xy.x+size/2), Math.max(0, xy.y-size/2),
				Math.min(h, xy.y+size/2+1));	
			if(foundEdge) break;
			size+=2;
		}

		int min=2*size*size;
		// Go through the area and find the edge pixel closest to 'xy'.
		for(int i=Math.max(0, xy.y-size/2); i<Math.min(h, xy.y+size/2+1); i++){
			for(int j=Math.max(0, xy.x-size/2); j<Math.min(w, xy.x+size/2+1); j++){
				if(logzx[i][j]>0 && (i-xy.y)*(i-xy.y)+(j-xy.x)*(j-xy.x)<min){
					min=(i-xy.y)*(i-xy.y)+(j-xy.x)*(j-xy.x);
					ret.x=j;
					ret.y=i;
				}
			}
		}
		
		return ret;
	}
	
	/** Go through row 'y' in 'logxz' from 'x' to 'bound' looking for an edge pixel.
	 * @param x X coordinate to start looking from.
	 * @param bound Upper bound on x.
	 * @param y Y coordinate of the row.
	 * @return true if there's an edge pixel, false if not.
	 */
	private boolean isThereEdgeH(int x, int bound, int y){
		for(; x<bound; x++)
			if(logzx[y][x]>0)
				return true;
		return false;
	}
	
	/** Go through column 'x' in 'logxz' from 'y' to 'bound' looking for an edge pixel.
	 * @param x X coordinate of the column.
	 * @param y Y coordinate to start looking from.
	 * @param bound Upper bound on y.
	 * @return true if there's an edge pixel, false if not.
	 */
	private boolean isThereEdgeV(int x, int y, int bound){
		for(; y<bound; y++)
			if(logzx[y][x]>0)
				return true;
		return false;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mousePressed(java.awt.event.MouseEvent)
	 */
	public boolean mousePressed(MouseEvent e) {
		return false;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mouseReleased(java.awt.event.MouseEvent)
	 */
	public boolean mouseReleased(MouseEvent e) {
		return false;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mouseEntered(java.awt.event.MouseEvent)
	 */
	public boolean mouseEntered(MouseEvent e) {
		imgPanel.setCursor(Laf.getCursor("scissors"));
		return false;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mouseExited(java.awt.event.MouseEvent)
	 */
	public boolean mouseExited(MouseEvent e) {
		imgPanel.setCursor(Laf.getCursor("default"));
		return false;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mouseClicked(java.awt.event.MouseEvent)
	 */
	public boolean mouseClicked(MouseEvent e) {
		if(logzx==null && !initialize())
			return false;
		XYPair xy;
		if(closedPath){
			// Start drawing a new path around an object.
			seeds.clear();
			path.clear();
			xy=snapToEdge(new XYPair(e.getX(), e.getY()));
			computePaths(xy);
			closedPath=false;
			seeds.add(xy);
		}else if(e.getClickCount()>1){
			// Finish drawing the path.
			closedPath=true;

			xy=snapToEdge(new XYPair(e.getX(), e.getY()));
			computePaths(xy);
			path.add(new Shape(getPath((XYPair)seeds.firstElement())));
		}else {
			// new seed
			xy=snapToEdge(new XYPair(e.getX(), e.getY()));
			Shape s=new Shape(getPath(xy));
			if(lastPathSeg<0) path.add(s);
			else path.set(lastPathSeg, s);
			seeds.add(xy);
			computePaths(xy);
		}
		path.add(new Circle(2, xy.x, xy.y));
		lastPathSeg=-1;
		dest=null;
		
		if(closedPath){
			Vector tmp, sPath=new Vector();
			int minX=image.getWidth(), minY=image.getHeight();
			
			// fill in pixels along the selection path.
			for(int i=0; i<path.size(); i++){
				if(path.elementAt(i) instanceof Circle)
					continue;

				tmp=((Shape)path.elementAt(i)).pos;
				for(int j=tmp.size()-1; j>=0; j--){
					xy=(XYPair)tmp.elementAt(j);
					if(xy.y<minY) minY=xy.y;
					if(xy.x<minX) minX=xy.x;
					sPath.add(xy);
				}
			}
			path.clear();
			path.add(new SelectionShape(sPath));
			imgPanel.setSelectionX(minX);
			imgPanel.setSelectionY(minY);
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mouseMoved(java.awt.event.MouseEvent)
	 */
	public boolean mouseMoved(MouseEvent e) {
		dest=new XYPair(e.getX(), e.getY());
		return true;
	}

	/* (non-Javadoc)
	 * @see ICOperation#mouseDragged(java.awt.event.MouseEvent)
	 */
	public boolean mouseDragged(MouseEvent e) {
		return false;
	}

	/* (non-Javadoc)
	 * @see ICOperation#beforeDrawing()
	 */
	public void beforeDrawing(){
		if(!closedPath && dest!=null){
			Shape p=new Shape(getPath(snapToEdge(dest)));
			if(lastPathSeg<0){
				path.add(p);
				lastPathSeg=path.size()-1;
			}else
				path.set(lastPathSeg, p);
		}
	}
	
	/**
	 * Class for running initalization operations.
	 */
	private class RunInit extends ProcessesLump{
		double sigma=-1;
		
		/* (non-Javadoc)
		 * @see ProcessesLump#precompute()
		 */
		protected boolean precompute(){
			String s;
			// Keep asking the user for the sigma until a valid value is entered.
			while (sigma<0) {
				s = (String) JOptionPane.showInputDialog(imgPanel,
						"Enter the sigma value to be\n used in the Gaussian mask:",
						"Sigma", JOptionPane.PLAIN_MESSAGE, null, null, "1");
					
				if(s==null) return false;
			
				try {
					sigma = Double.parseDouble(s);
					break;
				} catch (Exception e) {
				}
			}
			return true;
		}

		/* (non-Javadoc)
		 * @see ProcessesLump#estimateProgress()
		 */
		protected int estimateProgress(){
			return Paint.estimateGetZeroXingsProgress(image, sigma)+
					Paint.estimateGetGradientProgress(image);
		}

		/* (non-Javadoc)
		 * @see ProcessesLump#exec()
		 */
		protected Object exec(){
			logzx=Paint.getZeroXings(imgPanel, image, sigma);
			grad=Paint.getGradient(image);
			return logzx;
		}
	}
}
