import cern.colt.matrix.impl.*;
import cern.colt.matrix.*;
import cern.colt.matrix.linalg.*;

// assuming that colt.jar is in the same directory...
// javac -classpath .:colt.jar TextureMap.java 
// java -cp .:colt.jar TextureMap

/** Class doing the warping algebra. */
public class TextureMap {

	private Algebra a=new Algebra();
	private DoubleMatrix2D mapMatrix; // tranformation

	/**
	 * Computes the source x coordinate.
	 * @param r X coordinate in the warped image.
	 * @param c Y coordinate in the warped image.
	 * @return Corresponding x coordinate in the source image.
	 */
	public double texU(double r, double c){
		double [] x1={r, c, 1};
		DoubleMatrix1D x=new DenseDoubleMatrix1D(x1);
		DoubleMatrix1D y=a.mult(mapMatrix, x);
		return y.get(0)/y.get(2);
	}

	/**
	 * Computes the source y coordinate.
	 * @param r X coordinate in the warped image.
	 * @param c Y coordinate in the warped image.
	 * @return Corresponding y coordinate in the source image.
	 */
	public double texV(double r, double c){
		double [] x1={r, c, 1};
		DoubleMatrix1D x=new DenseDoubleMatrix1D(x1);
		DoubleMatrix1D y=a.mult(mapMatrix, x);
		return y.get(1)/y.get(2);
	}

	/**
	 * Constructor:
	 * @param u1 x value of the source point 1.
	 * @param v1 y value of the source point 1.
	 * @param u2 x value of the source point 2.
	 * @param v2 y value of the source point 2.
	 * @param u3 x value of the source point 3.
	 * @param v3 y value of the source point 3.
	 * @param u4 x value of the source point 4.
	 * @param v4 y value of the source point 4.
	 * @param r1 x value of the target point 1.
	 * @param c1 y value of the target point 1.
	 * @param r2 x value of the target point 2.
	 * @param c2 y value of the target point 2.
	 * @param r3 x value of the target point 3.
	 * @param c3 y value of the target point 3.
	 * @param r4 x value of the target point 4.
	 * @param c4 y value of the target point 4.
	 */
	public TextureMap(
		int u1, int v1, int u2, int v2, int u3, int v3, int u4, int v4, 
		int r1, int c1, int r2, int c2, int r3, int c3, int r4, int c4
	){
		// linear system to solve to get the ais.
		double [][] d1={
			{ -r1, -c1, -1,  0,   0,  0, u1*r1, u1*c1 },
			{ 0  ,   0,  0,-r1, -c1, -1, v1*r1, v1*c1 },
			{ -r2, -c2, -1,  0,   0,  0, u2*r2, u2*c2 },
			{ 0  ,   0,  0,-r2, -c2, -1, v2*r2, v2*c2 },
			{ -r3, -c3, -1,  0,   0,  0, u3*r3, u3*c3 },
			{ 0  ,   0,  0,-r3, -c3, -1, v3*r3, v3*c3 },
			{ -r4, -c4, -1,  0,   0,  0, u4*r4, u4*c4 },
			{ 0  ,   0,  0,-r4, -c4, -1, v4*r4, v4*c4 }
		};
		DoubleMatrix2D A=new DenseDoubleMatrix2D(d1);
		// the solution vector.
		double [][] d2= { { -u1 }, { -v1 } , { -u2 } , { -v2 } , { -u3 } , { -v3 } , { -u4 } , { -v4 } };
		DoubleMatrix2D b=new DenseDoubleMatrix2D(d2);
		DoubleMatrix2D s=a.solve(A,b);
		// create the tranformation.
		double [][] t1={
			{ s.get(0,0), s.get(1,0), s.get(2,0) },
			{ s.get(3,0), s.get(4,0), s.get(5,0) },
			{ s.get(6,0), s.get(7,0), 1  },

		};
		mapMatrix=new DenseDoubleMatrix2D(t1);
	}
	
	/**
	 * Converts the tranformation to a string.
	 * @return String representation of the transformation.
	 */
	public String toString(){ return mapMatrix.toString(); }
}
