import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** Panel holding the image to be warped. */
class WarpSourcePanel extends ImagePanel implements MouseListener, MouseMotionListener {
	private Laf icApp;
	private SelectorPoint [] path;	// 4 source selector points.
	private SelectorPoint drag;		// selector being dragged

	/** Constructor: 
	 * @param filename The path to the image file.
	 * @param icApp The main ImageCompositing application.
	 */
	public WarpSourcePanel(String filename, Laf icApp){
		super(filename);
		this.icApp=icApp;

		// set up keyboard events to listen to arrows keys.
		InputMap im = getInputMap();
		ActionMap am = getActionMap();
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), "down");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_DOWN, 0), "down");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), "up");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_UP, 0), "up");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_LEFT, 0), "left");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "left");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_RIGHT, 0), "right");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "right");

		// event handlers for keys, perform nudges to the selectors.
		am.put("down", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.y++;
				repaint();
			}
		});
		am.put("up", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.y--;
				repaint();
			}
		});
		am.put("left", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.x--;
				repaint();
			}
		});
		am.put("right", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				if (drag == null)
					return;
				drag.center.x++;
				repaint();
			}
		});
		
		SelectorPoint [] tmp={
			new SelectorPoint(3, 3, 3, Color.BLUE),
			new SelectorPoint(3, 3, getImageHeight()-4, Color.RED),
			new SelectorPoint(3, getImageWidth()-4, 3, Color.GREEN),
			new SelectorPoint(3, getImageWidth()-4, getImageHeight()-4, Color.YELLOW)
		};
		path=tmp;
	}
	
	/**
	 * returns the array of circles.
	 * @return Array of selector points.
	 */
	public SelectorPoint [] getPath(){
		return path;
	}

	/** Event handler for mouse presses.
	 * @param e Event object describing the mouse event.
	 */
	public void mousePressed(MouseEvent e) {
		super.mousePressed(e);
		if(getState()!=POINTER) return;
		int x=e.getX(), y=e.getY();
		if (drag != null) {
			drag.selected = false;
			drag = null;
		}
		for (int i = 0; i < path.length; i++) {
			if (path[i].hitTest(x, y)) {
				drag = path[i];
				drag.selected = true;
				break;
			}
		}
		repaint();
	}
	
	/** Event handler for the mouse move event.
	 * @param e Object describing the event.
	 */
	public void mouseMoved(MouseEvent e){
		super.mouseMoved(e);
		if(getState()!=POINTER) return;
		int x=e.getX(), y=e.getY();
		for(int i=0; i<path.length; i++){
			if(path[i].hitTest(x, y)){
				setCursor(new Cursor(Cursor.HAND_CURSOR));
				return;
			}
		}
		setCursor(Cursor.getDefaultCursor());
	}
	
	/** Event handler for the mouse drag event.
	 * @param e Object describing the event.
	 */
	public void mouseDragged(MouseEvent e){
		super.mouseDragged(e);
		if(drag==null || getState()!=POINTER) return;
		drag.center.x=e.getX();
		drag.center.y=e.getY();
		repaint();
	}

	/** Paints this component.
	 * @param g Graphics object to paint on.
	 */
	public void paintComponent(Graphics g){
		super.paintComponent(g);

		for(int i=0; path!=null && i<path.length; i++)
			path[i].paint(g);
	}
}
